"""Modules for Inferer
It include impleentation of the Elo and Glicko rating system
"""

from .elo import Elo
from .glicko import Glicko
from .playerlevel import PlayerLevel
from .playerwinprc import PlayerWinPRC
from .eventscoring import EventScoring
